@@ -40,7 +40,7 @@ def tokenize(*args, **kwargs):
     True
     """
     if kwargs:
-        args = args + (kwargs,)
+        args += (kwargs,)
     return md5(str(tuple(args)).encode()).hexdigest()
 
 
@@ -65,6 +65,7 @@ def split_path(path):
     else:
         return path.split('/', 1)
 
+
 key_acls = {'private', 'public-read', 'public-read-write',
             'authenticated-read', 'aws-exec-read', 'bucket-owner-read',
             'bucket-owner-full-control'}
@@ -98,8 +99,8 @@ class S3FileSystem(object):
     use_ssl : bool (True)
         Whether to use SSL in connections to S3; may be faster without, but
         insecure
-    s3_additional_kwargs : dict of parameters that are used when calling s3 api methods.
-        Typically used for things like "ServerSideEncryption".
+    s3_additional_kwargs : dict of parameters that are used when calling s3 api
+        methods. Typically used for things like "ServerSideEncryption".
     client_kwargs : dict of parameters for the boto3 client
     requester_pays : bool (False)
         If RequesterPays buckets are supported.
@@ -109,6 +110,10 @@ class S3FileSystem(object):
     default_fill_cache : Bool (True)
         Whether to use cache filling with open by default. Refer to
         ``S3File.open``.
+    version_aware : bool (False)
+        Whether to support bucket versioning.  If enable this will require the
+        user to have the neccesary IAM permissions for dealing with versioned
+        objects.
     config_kwargs : dict of parameters passed to ``botocore.client.Config``
     kwargs : other parameters for boto3 session
 
@@ -131,7 +136,8 @@ class S3FileSystem(object):
     def __init__(self, anon=False, key=None, secret=None, token=None,
                  use_ssl=True, client_kwargs=None, requester_pays=False,
                  default_block_size=None, default_fill_cache=True,
-                 config_kwargs=None, s3_additional_kwargs=None, **kwargs):
+                 version_aware=False, config_kwargs=None,
+                 s3_additional_kwargs=None, **kwargs):
         self.anon = anon
         self.session = None
         self.key = key
@@ -146,6 +152,7 @@ class S3FileSystem(object):
         if config_kwargs is None:
             config_kwargs = {}
         self.default_fill_cache = default_fill_cache
+        self.version_aware = version_aware
         self.client_kwargs = client_kwargs
         self.config_kwargs = config_kwargs
         self.dirs = {}
@@ -160,7 +167,8 @@ class S3FileSystem(object):
         return self._kwargs_helper.filter_dict(s3_method.__name__, kwargs)
 
     def _call_s3(self, method, *akwarglist, **kwargs):
-        additional_kwargs = self._get_s3_method_kwargs(method, *akwarglist, **kwargs)
+        additional_kwargs = self._get_s3_method_kwargs(method, *akwarglist,
+                                                       **kwargs)
         return method(**additional_kwargs)
 
     def _get_s3_method_kwargs(self, method, *akwarglist, **kwargs):
@@ -259,11 +267,10 @@ class S3FileSystem(object):
 
     def __setstate__(self, state):
         self.__dict__.update(state)
-        self._conn = {}
         self.s3 = self.connect()
         self._kwargs_helper = ParamKwargsHelper(self.s3)
 
-    def open(self, path, mode='rb', block_size=None, acl='',
+    def open(self, path, mode='rb', block_size=None, acl='', version_id=None,
              fill_cache=None, **kwargs):
         """ Open a file for reading or writing
 
@@ -282,8 +289,13 @@ class S3FileSystem(object):
             out of a file, performance may be better if False.
         acl: str
             Canned ACL to set when writing
+        version_id : str
+            Explicit version of the object to open.  This requires that the s3
+            filesystem is version aware and bucket versioning is enabled on the
+            relavant bucket.
         kwargs: dict-like
-            Additional parameters used for s3 methods.  Typically used for ServerSideEncryption.
+            Additional parameters used for s3 methods.  Typically used for
+            ServerSideEncryption.
         """
         if block_size is None:
             block_size = self.default_block_size
@@ -295,8 +307,12 @@ class S3FileSystem(object):
         acl = acl or self.s3_additional_kwargs.get('ACL', '')
         kw = self.s3_additional_kwargs.copy()
         kw.update(kwargs)
+        if not self.version_aware and version_id:
+            raise ValueError("version_id cannot be specified if the filesystem "
+                             "is not version aware")
         return S3File(self, path, mode, block_size=block_size, acl=acl,
-                      fill_cache=fill_cache, s3_additional_kwargs=kw)
+                      version_id=version_id, fill_cache=fill_cache,
+                      s3_additional_kwargs=kw)
 
     def _lsdir(self, path, refresh=False):
         if path.startswith('s3://'):
@@ -352,13 +368,8 @@ class S3FileSystem(object):
         ----------
         path : string/bytes
             location at which to list files
-        detail : bool (=True)
-            if True, each list item is a dict of file properties;
-            otherwise, returns list of filenames
         refresh : bool (=False)
             if False, look in local cache for file details first
-        kwargs : dict
-            additional arguments passed on
         """
         if path.startswith('s3://'):
             path = path[len('s3://'):]
@@ -368,7 +379,20 @@ class S3FileSystem(object):
             return self._lsdir(path, refresh)
 
     def ls(self, path, detail=False, refresh=False, **kwargs):
-        """ List single "directory" with or without details """
+        """ List single "directory" with or without details
+
+        Parameters
+        ----------
+        path : string/bytes
+            location at which to list files
+        detail : bool (=True)
+            if True, each list item is a dict of file properties;
+            otherwise, returns list of filenames
+        refresh : bool (=False)
+            if False, look in local cache for file details first
+        kwargs : dict
+            additional arguments passed on
+        """
         if path.startswith('s3://'):
             path = path[len('s3://'):]
         path = path.rstrip('/')
@@ -383,35 +407,68 @@ class S3FileSystem(object):
         else:
             return [f['Key'] for f in files]
 
-    def info(self, path, **kwargs):
+    def info(self, path, version_id=None, refresh=False, **kwargs):
         """ Detail on the specific file pointed to by path.
 
         Gets details only for a specific key, directories/buckets cannot be
         used with info.
+
+        Parameters
+        ----------
+        version_id : str, optional
+            version of the key to perform the head_object on
+        refresh : bool
+            If true, don't look in the info cache
         """
         parent = path.rsplit('/', 1)[0]
 
-        if path in self.dirs:
-            files = self.dirs[path]
-            if len(files) == 1:
-                return files[0]
-        elif parent in self.dirs:
-            for f in self.dirs[parent]:
-                if f['Key'] == path:
-                    return f
+        if not refresh:
+            if path in self.dirs:
+                files = self.dirs[path]
+                if len(files) == 1:
+                    return files[0]
+            elif parent in self.dirs:
+                for f in self.dirs[parent]:
+                    if f['Key'] == path:
+                        return f
 
         try:
             bucket, key = split_path(path)
+            if version_id is not None:
+                if not self.version_aware:
+                    raise ValueError("version_id cannot be specified if the "
+                                     "filesystem is not version aware")
+                kwargs['VersionId'] = version_id
             out = self._call_s3(self.s3.head_object, kwargs, Bucket=bucket,
                                 Key=key, **self.req_kw)
-            out = {'ETag': out['ETag'], 'Key': '/'.join([bucket, key]),
-                    'LastModified': out['LastModified'],
-                    'Size': out['ContentLength'], 'StorageClass': "STANDARD"}
+            out = {
+                'ETag': out['ETag'],
+                'Key': '/'.join([bucket, key]),
+                'LastModified': out['LastModified'],
+                'Size': out['ContentLength'],
+                'StorageClass': "STANDARD",
+                'VersionId': out.get('VersionId')
+            }
             return out
-        except (ClientError, ParamValidationError) as e:
+        except (ClientError, ParamValidationError):
             logger.debug("Failed to head path %s", path, exc_info=True)
             raise FileNotFoundError(path)
 
+    def object_version_info(self, path, **kwargs):
+        if not self.version_aware:
+            raise ValueError("version specific functionality is disabled for "
+                             "non-version aware filesystems")
+        bucket, key = split_path(path)
+        kwargs = {}
+        out = {'IsTruncated': True}
+        versions = []
+        while out['IsTruncated']:
+            out = self._call_s3(self.s3.list_object_versions, kwargs,
+                                Bucket=bucket, Prefix=key, **self.req_kw)
+            versions.extend(out['Versions'])
+            kwargs['VersionIdMarker'] = out.get('NextVersionIdMarker', '')
+        return versions
+
     _metadata_cache = {}
 
     def metadata(self, path, refresh=False, **kwargs):
@@ -468,14 +525,16 @@ class S3FileSystem(object):
         mode:
             One of 'o' or 'm'
             'o': Will over-write any existing tags.
-            'm': Will merge in new tags with existing tags.  Incurs two remote calls.
+            'm': Will merge in new tags with existing tags.  Incurs two remote
+            calls.
         """
         bucket, key = split_path(path)
 
         if mode == 'm':
             existing_tags = self.get_tags(path=path)
             existing_tags.update(tags)
-            new_tags = [{'Key': k, 'Value': v} for k, v in existing_tags.items()]
+            new_tags = [{'Key': k, 'Value': v}
+                        for k, v in existing_tags.items()]
         elif mode == 'o':
             new_tags = [{'Key': k, 'Value': v} for k, v in tags.items()]
         else:
@@ -501,20 +560,25 @@ class S3FileSystem(object):
     def setxattr(self, path, copy_kwargs=None, **kw_args):
         """ Set metadata.
 
-        Attributes have to be of the form documented in the `Metadata Reference`_.
+        Attributes have to be of the form documented in the
+        `Metadata Reference`_.
 
         Parameters
         ---------
-        kw_args : key-value pairs like field="value", where the values must be strings. Does not alter existing fields,
-            unless the field appears here - if the value is None, delete the field.
-        copy_args : dict, optional
-            dictionary of additional params to use for the underlying s3.copy_object.
+        kw_args : key-value pairs like field="value", where the values must be
+            strings. Does not alter existing fields, unless
+            the field appears here - if the value is None, delete the
+            field.
+        copy_kwargs : dict, optional
+            dictionary of additional params to use for the underlying
+            s3.copy_object.
 
         Examples
         --------
         >>> mys3file.setxattr(attribute_1='value1', attribute_2='value2')  # doctest: +SKIP
         # Example for use with copy_args
-        >>> mys3file.setxattr(copy_kwargs={'ContentType': 'application/pdf'}, attribute_1='value1')  # doctest: +SKIP
+        >>> mys3file.setxattr(copy_kwargs={'ContentType': 'application/pdf'},
+        ...     attribute_1='value1')  # doctest: +SKIP
 
 
         .. Metadata Reference:
@@ -544,7 +608,7 @@ class S3FileSystem(object):
         # refresh metadata
         self._metadata_cache[path] = metadata
 
-    def _walk(self, path, refresh=False):
+    def _walk(self, path, refresh=False, directories=False):
         if path.startswith('s3://'):
             path = path[len('s3://'):]
         if path in ['', '/']:
@@ -554,11 +618,11 @@ class S3FileSystem(object):
             if f['StorageClass'] == 'DIRECTORY':
                 filenames.extend(self._walk(f['Key'], refresh))
         return [f for f in filenames if f['StorageClass'] not in
-                ['BUCKET', 'DIRECTORY']]
+                (['BUCKET'] + ['DIRECTORY'] if not directories else [])]
 
-    def walk(self, path, refresh=False):
+    def walk(self, path, refresh=False, directories=False):
         """ Return all real keys below path """
-        return [f['Key'] for f in self._walk(path, refresh)]
+        return [f['Key'] for f in self._walk(path, refresh, directories)]
 
     def chmod(self, path, acl, **kwargs):
         """ Set Access Control on a bucket/key
@@ -597,23 +661,31 @@ class S3FileSystem(object):
         bucket, key = split_path(path)
         if "*" in bucket:
             raise ValueError('Bucket cannot contain a "*"')
-        if '*' not in path:
-            path = path.rstrip('/') + '/*'
-        if '/' in path[:path.index('*')]:
-            ind = path[:path.index('*')].rindex('/')
-            root = path[:ind + 1]
-        else:
-            root = '/'
-        allfiles = self.walk(root)
-        pattern = re.compile("^" + path.replace('//', '/')
-                             .rstrip('/')
-                             .replace('*', '[^/]*')
-                             .replace('?', '.') + "$")
-        out = [f for f in allfiles if re.match(pattern,
-               f.replace('//', '/').rstrip('/'))]
-        if not out:
-            out = self.ls(path0)
-        return out
+
+        # A slash (/) is guaranteed to exist before the first occurrence of
+        # star (*) in the path: since the * is NOT in the bucket name, there
+        # must be a / between the bucket name and the *.
+        star_pos = path.find('*')
+        root = path if star_pos == -1 else path[:path[:star_pos].rindex('/') + 1]
+        allfiles = self.walk(root, directories=True)
+
+        # Translate the glob pattern into regex (similar to `fnmatch`).
+        regex_text = '('
+        for c in path:
+            if c == '*':
+                regex_text += '[^/]*'
+            elif c == '?':
+                regex_text += '.'
+            else:
+                regex_text += re.escape(c)
+        regex_text += ')'
+
+        regex_pattern = re.compile(regex_text)
+
+        seen = set()
+        return [p for f in allfiles
+                      for m in [regex_pattern.match(f)] if m
+                          for p in [m.group(0)] if not (p in seen or seen.add(p))]
 
     def du(self, path, total=False, deep=False, **kwargs):
         """ Bytes in keys at path """
@@ -665,9 +737,10 @@ class S3FileSystem(object):
             the number of seconds this signature will be good for.
         """
         bucket, key = split_path(path)
-        return self.s3.generate_presigned_url(ClientMethod='get_object',
-                                       Params=dict(Bucket=bucket, Key=key, **kwargs),
-                                       ExpiresIn=expires)
+        return self.s3.generate_presigned_url(
+            ClientMethod='get_object', Params=dict(Bucket=bucket, Key=key,
+                                                   **kwargs),
+            ExpiresIn=expires)
 
     def get(self, path, filename, **kwargs):
         """ Stream data from file at path to local filename """
@@ -700,7 +773,8 @@ class S3FileSystem(object):
             path = path[len('s3://'):]
         path = path.rstrip('/')
         bucket, key = split_path(path)
-        if not self._ls(path) and ((key and self.info(path)['Size'] == 0) or not key):
+        if not self._ls(path) and ((key and self.info(path)['Size'] == 0)
+                                   or not key):
             self.rm(path, **kwargs)
         else:
             raise IOError('Path is not directory-like', path)
@@ -731,14 +805,17 @@ class S3FileSystem(object):
             Key=key
             )
         out = [self._call_s3(
-                    self.s3.upload_part_copy,
-                    kwargs,
-                    Bucket=bucket, Key=key, UploadId=mpu['UploadId'],
-                    CopySource=f, PartNumber=i+1) for (i, f) in enumerate(filelist)]
-        parts = [{'PartNumber': i+1, 'ETag': o['CopyPartResult']['ETag']} for (i, o) in enumerate(out)]
+            self.s3.upload_part_copy,
+            kwargs,
+            Bucket=bucket, Key=key, UploadId=mpu['UploadId'],
+            CopySource=f, PartNumber=i + 1)
+            for (i, f) in enumerate(filelist)]
+        parts = [{'PartNumber': i + 1, 'ETag': o['CopyPartResult']['ETag']} for
+                 (i, o) in enumerate(out)]
         part_info = {'Parts': parts}
         self.s3.complete_multipart_upload(Bucket=bucket, Key=key,
-                    UploadId=mpu['UploadId'], MultipartUpload=part_info)
+                                          UploadId=mpu['UploadId'],
+                                          MultipartUpload=part_info)
         self.invalidate_cache(path)
 
     def copy_basic(self, path1, path2, **kwargs):
@@ -791,13 +868,14 @@ class S3FileSystem(object):
             return
         buckets = {split_path(path)[0] for path in pathlist}
         if len(buckets) > 1:
-            raise ValueError("Bulk delete files should refer to only one bucket")
+            raise ValueError("Bulk delete files should refer to only one "
+                             "bucket")
         bucket = buckets.pop()
         if len(pathlist) > 1000:
             for i in range((len(pathlist) // 1000) + 1):
                 self.bulk_delete(pathlist[i*1000:(i+1)*1000])
             return
-        delete_keys = {'Objects': [{'Key' : split_path(path)[1]} for path
+        delete_keys = {'Objects': [{'Key': split_path(path)[1]} for path
                                    in pathlist]}
         try:
             self._call_s3(
@@ -923,8 +1001,8 @@ class S3FileSystem(object):
                 length = size
             if offset + length > size:
                 length = size - offset
-            bytes = read_block(f, offset, length, delimiter)
-        return bytes
+            b = read_block(f, offset, length, delimiter)
+        return b
 
 
 class S3File(object):
@@ -933,18 +1011,23 @@ class S3File(object):
 
     Parameters
     ----------
-    s3 : boto3 connection
+    s3 : S3FileSystem
+        boto3 connection
     path : string
         S3 bucket/key to access
     block_size : int
         read-ahead size for finding delimiters
-    fill_cache: bool
+    fill_cache : bool
         If seeking to new a part of the file beyond the current buffer,
         with this True, the buffer will be filled between the sections to
         best support random access. When reading only a few specific chunks
         out of a file, performance may be better if False.
     acl: str
         Canned ACL to apply
+    version_id : str
+        Optional version to read the file at.  If not specified this will
+        default to the current version of the object.  This is only used for
+        reading.
 
     Examples
     --------
@@ -958,10 +1041,11 @@ class S3File(object):
     """
 
     def __init__(self, s3, path, mode='rb', block_size=5 * 2 ** 20, acl="",
-                 fill_cache=True, s3_additional_kwargs=None):
+                 version_id=None, fill_cache=True, s3_additional_kwargs=None):
         self.mode = mode
         if mode not in {'rb', 'wb', 'ab'}:
-            raise NotImplementedError("File mode must be {'rb', 'wb', 'ab'}, not %s" % mode)
+            raise NotImplementedError("File mode must be {'rb', 'wb', 'ab'}, "
+                                      "not %s" % mode)
         if path.startswith('s3://'):
             path = path[len('s3://'):]
         self.path = path
@@ -977,6 +1061,7 @@ class S3File(object):
         self.closed = False
         self.trim = True
         self.mpu = None
+        self.version_id = version_id
         self.acl = acl
         self.fill_cache = fill_cache
         self.s3_additional_kwargs = s3_additional_kwargs or {}
@@ -1014,7 +1099,10 @@ class S3File(object):
                                        'ETag': out['CopyPartResult']['ETag']})
         else:
             try:
-                self.size = self.info()['Size']
+                info = self.info()
+                self.size = info['Size']
+                if self.s3.version_aware:
+                    self.version_id = info.get('VersionId')
             except (ClientError, ParamValidationError):
                 raise IOError("File not accessible", path)
 
@@ -1024,7 +1112,12 @@ class S3File(object):
 
     def info(self, **kwargs):
         """ File information about this path """
-        return self.s3.info(self.path, **kwargs)
+        # When the bucket is version aware we need to explicitly get the
+        # correct filesize and for the particular version. In the case where
+        # self.version_id is None, this will be the most recent version.
+        refresh = self.s3.version_aware
+        return self.s3.info(self.path, version_id=self.version_id,
+                            refresh=refresh, **kwargs)
 
     def metadata(self, refresh=False, **kwargs):
         """ Return metadata of file.
@@ -1054,7 +1147,8 @@ class S3File(object):
         >>> mys3file.setxattr(attribute_1='value1', attribute_2='value2')  # doctest: +SKIP
         """
         if self.mode != 'rb':
-            raise NotImplementedError('cannot update metadata while file is open for writing')
+            raise NotImplementedError('cannot update metadata while file '
+                                      'is open for writing')
         return self.s3.setxattr(self.path, copy_kwargs=copy_kwargs, **kwargs)
 
     def url(self, **kwargs):
@@ -1101,7 +1195,7 @@ class S3File(object):
         self._fetch(self.loc, self.loc + 1)
         while True:
             found = self.cache[self.loc - self.start:].find(b'\n') + 1
-            if length > 0 and found > length:
+            if 0 < length < found:
                 return self.read(length)
             if found:
                 return self.read(found)
@@ -1125,17 +1219,24 @@ class S3File(object):
         return list(self)
 
     def _fetch(self, start, end):
+        # if we have not enabled version_aware then we should use the
+        # latest version.
+        if self.s3.version_aware:
+            version_id = self.version_id
+        else:
+            version_id = None
         if self.start is None and self.end is None:
             # First read
             self.start = start
             self.end = end + self.blocksize
             self.cache = _fetch_range(self.s3.s3, self.bucket, self.key,
-                                      start, self.end, req_kw=self.s3.req_kw)
+                                      version_id, start, self.end,
+                                      req_kw=self.s3.req_kw)
         if start < self.start:
             if not self.fill_cache and end + self.blocksize < self.start:
                 self.start, self.end = None, None
                 return self._fetch(start, end)
-            new = _fetch_range(self.s3.s3, self.bucket, self.key,
+            new = _fetch_range(self.s3.s3, self.bucket, self.key, version_id,
                                start, self.start, req_kw=self.s3.req_kw)
             self.start = start
             self.cache = new + self.cache
@@ -1145,7 +1246,7 @@ class S3File(object):
             if not self.fill_cache and start > self.end:
                 self.start, self.end = None, None
                 return self._fetch(start, end)
-            new = _fetch_range(self.s3.s3, self.bucket, self.key,
+            new = _fetch_range(self.s3.s3, self.bucket, self.key, version_id,
                                self.end, end + self.blocksize,
                                req_kw=self.s3.req_kw)
             self.end = end + self.blocksize
@@ -1181,7 +1282,8 @@ class S3File(object):
         """
         Write data to buffer.
 
-        Buffer only sent to S3 on close() or if buffer is greater than or equal to blocksize.
+        Buffer only sent to S3 on close() or if buffer is greater than or equal
+        to blocksize.
 
         Parameters
         ----------
@@ -1213,6 +1315,7 @@ class S3File(object):
         force : bool
             When closing, write the last block even if it is smaller than
             blocks are allowed to be.
+        retries: int
         """
         if self.mode in {'wb', 'ab'} and not self.closed:
             if self.buffer.tell() < self.blocksize and not force:
@@ -1319,16 +1422,20 @@ class S3File(object):
         self.close()
 
 
-def _fetch_range(client, bucket, key, start, end, max_attempts=10,
+def _fetch_range(client, bucket, key, version_id, start, end, max_attempts=10,
                  req_kw=None):
     if req_kw is None:
         req_kw = {}
     logger.debug("Fetch: %s/%s, %s-%s", bucket, key, start, end)
     for i in range(max_attempts):
         try:
+            if version_id is not None:
+                kwargs = dict({'VersionId': version_id}, **req_kw)
+            else:
+                kwargs = req_kw
             resp = client.get_object(Bucket=bucket, Key=key,
                                      Range='bytes=%i-%i' % (start, end - 1),
-                                     **req_kw)
+                                     **kwargs)
             return resp['Body'].read()
         except S3_RETRYABLE_ERRORS as e:
             logger.debug('Exception %e on S3 download, retrying', e,
