@@ -6,7 +6,6 @@ import re
 import time
 import pytest
 from itertools import chain
-from os.path import join
 from s3fs.core import S3FileSystem
 from s3fs.utils import seek_delimiter, ignoring, tmpfile, SSEParams
 import moto
@@ -15,6 +14,7 @@ from botocore.exceptions import NoCredentialsError
 
 test_bucket_name = 'test'
 secure_bucket_name = 'test-secure'
+versioned_bucket_name = 'test-versioned'
 files = {'test/accounts.1.json':  (b'{"amount": 100, "name": "Alice"}\n'
                                    b'{"amount": 200, "name": "Bob"}\n'
                                    b'{"amount": 300, "name": "Charlie"}\n'
@@ -37,6 +37,8 @@ text_files = {'nested/file1': b'hello\n',
               'nested/file2': b'world',
               'nested/nested2/file1': b'hello\n',
               'nested/nested2/file2': b'world'}
+glob_files = {'file.dat': b'',
+              'filexdat': b''}
 a = test_bucket_name+'/tmp/test/a'
 b = test_bucket_name+'/tmp/test/b'
 c = test_bucket_name+'/tmp/test/c'
@@ -46,57 +48,63 @@ d = test_bucket_name+'/tmp/test/d'
 @pytest.yield_fixture
 def s3():
     # writable local S3 system
-    m = moto.mock_s3()
-    m.start()
-    import boto3
-    client = boto3.client('s3')
-    client.create_bucket(Bucket=test_bucket_name, ACL='public-read')
+    try:
+        m = moto.mock_s3()
+        m.start()
+        import boto3
+        client = boto3.client('s3')
+        client.create_bucket(Bucket=test_bucket_name, ACL='public-read')
 
-    # initialize secure bucket
-    bucket = client.create_bucket(Bucket=secure_bucket_name, ACL='public-read')
-    policy = json.dumps({
-        "Version": "2012-10-17",
-        "Id": "PutObjPolicy",
-        "Statement": [
-            {
-                "Sid": "DenyUnEncryptedObjectUploads",
-                "Effect": "Deny",
-                "Principal": "*",
-                "Action": "s3:PutObject",
-                "Resource": "arn:aws:s3:::{bucket_name}/*".format(bucket_name=secure_bucket_name),
-                "Condition": {
-                    "StringNotEquals": {
-                        "s3:x-amz-server-side-encryption": "aws:kms"
+        bucket = client.create_bucket(Bucket=versioned_bucket_name, ACL='public-read')
+        bucket_versioning = boto3.resource('s3').BucketVersioning(versioned_bucket_name)
+        bucket_versioning.enable()
+
+        # initialize secure bucket
+        bucket = client.create_bucket(Bucket=secure_bucket_name, ACL='public-read')
+        policy = json.dumps({
+            "Version": "2012-10-17",
+            "Id": "PutObjPolicy",
+            "Statement": [
+                {
+                    "Sid": "DenyUnEncryptedObjectUploads",
+                    "Effect": "Deny",
+                    "Principal": "*",
+                    "Action": "s3:PutObject",
+                    "Resource": "arn:aws:s3:::{bucket_name}/*".format(bucket_name=secure_bucket_name),
+                    "Condition": {
+                        "StringNotEquals": {
+                            "s3:x-amz-server-side-encryption": "aws:kms"
+                        }
                     }
                 }
-            }
-        ]
-    })
-    client.put_bucket_policy(Bucket=secure_bucket_name, Policy=policy)
+            ]
+        })
+        client.put_bucket_policy(Bucket=secure_bucket_name, Policy=policy)
 
-    for k in [a, b, c, d]:
-        try:
-            client.delete_object(Bucket=test_bucket_name, Key=k)
-        except:
-            pass
-    for flist in [files, csv_files, text_files]:
-        for f, data in flist.items():
-            client.put_object(Bucket=test_bucket_name, Key=f, Body=data)
-    yield S3FileSystem(anon=False)
-    for flist in [files, csv_files, text_files]:
-        for f, data in flist.items():
+        for k in [a, b, c, d]:
             try:
-                client.delete_object(Bucket=test_bucket_name, Key=f, Body=data)
-                client.delete_object(Bucket=secure_bucket_name, Key=f, Body=data)
+                client.delete_object(Bucket=test_bucket_name, Key=k)
             except:
                 pass
-    for k in [a, b, c, d]:
-        try:
-            client.delete_object(Bucket=test_bucket_name, Key=k)
-            client.delete_object(Bucket=secure_bucket_name, Key=k)
-        except:
-            pass
-    m.stop()
+        for flist in [files, csv_files, text_files, glob_files]:
+            for f, data in flist.items():
+                client.put_object(Bucket=test_bucket_name, Key=f, Body=data)
+        yield S3FileSystem(anon=False)
+        for flist in [files, csv_files, text_files, glob_files]:
+            for f, data in flist.items():
+                try:
+                    client.delete_object(Bucket=test_bucket_name, Key=f, Body=data)
+                    client.delete_object(Bucket=secure_bucket_name, Key=f, Body=data)
+                except:
+                    pass
+        for k in [a, b, c, d]:
+            try:
+                client.delete_object(Bucket=test_bucket_name, Key=k)
+                client.delete_object(Bucket=secure_bucket_name, Key=k)
+            except:
+                pass
+    finally:
+        m.stop()
 
 
 def test_simple(s3):
@@ -195,11 +203,13 @@ def test_xattr(s3):
     assert public_read_acl in s3.s3.get_object_acl(Bucket=bucket, Key=key)['Grants']
     assert s3.info(filename)['ETag'] == etag
 
+
 def test_xattr_setxattr_in_write_mode(s3):
     s3file = s3.open(a, 'wb')
     with pytest.raises(NotImplementedError):
         s3file.setxattr(test_xattr='1')
 
+
 @pytest.mark.xfail()
 def test_delegate(s3):
     out = s3.get_delegated_s3pars()
@@ -220,7 +230,7 @@ def test_not_delegate():
 
 
 def test_ls(s3):
-    assert set(s3.ls('')) == {test_bucket_name, secure_bucket_name}
+    assert set(s3.ls('')) == {test_bucket_name, secure_bucket_name, versioned_bucket_name}
     with pytest.raises((OSError, IOError)):
         s3.ls('nonexistent')
     fn = test_bucket_name+'/test/accounts.1.json'
@@ -231,6 +241,8 @@ def test_pickle(s3):
     import pickle
     s32 = pickle.loads(pickle.dumps(s3))
     assert s3.ls('test') == s32.ls('test')
+    s33 = pickle.loads(pickle.dumps(s32))
+    assert s3.ls('test') == s33.ls('test')
 
 
 def test_ls_touch(s3):
@@ -364,14 +376,23 @@ def test_s3_glob(s3):
     fn = test_bucket_name+'/nested/file1'
     assert fn not in s3.glob(test_bucket_name+'/')
     assert fn not in s3.glob(test_bucket_name+'/*')
-    assert fn in s3.glob(test_bucket_name+'/nested')
+    assert fn not in s3.glob(test_bucket_name+'/nested')
     assert fn in s3.glob(test_bucket_name+'/nested/*')
     assert fn in s3.glob(test_bucket_name+'/nested/file*')
     assert fn in s3.glob(test_bucket_name+'/*/*')
-    assert all(f in s3.walk(test_bucket_name) for f in s3.glob(test_bucket_name+'/nested/*'))
+    assert all(any(p.startswith(f + '/') or p == f
+                   for p in s3.walk(test_bucket_name, directories=True))
+                       for f in s3.glob(test_bucket_name+'/nested/*'))
+    assert [test_bucket_name + '/nested/nested2'] == s3.glob(test_bucket_name + '/nested/nested2')
+    assert [ 'test/nested/nested2/file1',
+             'test/nested/nested2/file2'] == s3.glob(test_bucket_name + '/nested/nested2/*')
+
     with pytest.raises(ValueError):
         s3.glob('*')
 
+    # Make sure glob() deals with the dot character (.) correctly.
+    assert test_bucket_name+'/file.dat' in s3.glob(test_bucket_name+'/file.*')
+    assert test_bucket_name+'/filexdat' not in s3.glob(test_bucket_name+'/file.*')
 
 def test_get_list_of_summary_objects(s3):
     L = s3.ls(test_bucket_name + '/test')
@@ -946,3 +967,51 @@ def test_tags(s3):
     s3.put_tags(fname, new_tagset, mode='m')
     tagset.update(new_tagset)
     assert s3.get_tags(fname) == tagset
+
+
+def test_versions(s3):
+    versioned_file = versioned_bucket_name + '/versioned_file'
+    s3 = S3FileSystem(anon=False, version_aware=True)
+    with s3.open(versioned_file, 'wb') as fo:
+        fo.write(b'1')
+    with s3.open(versioned_file, 'wb') as fo:
+        fo.write(b'2')
+    versions = s3.object_version_info(versioned_file)
+    assert len(versions) == 2
+
+    with s3.open(versioned_file) as fo:
+        assert fo.version_id == '1'
+        assert fo.read() == b'2'
+
+    with s3.open(versioned_file, version_id='0') as fo:
+        assert fo.version_id == '0'
+        assert fo.read() == b'1'
+
+
+def test_list_versions_many(s3):
+    # moto doesn't actually behave in the same way that s3 does here so this doesn't test
+    # anything really in moto 1.2
+    s3 = S3FileSystem(anon=False, version_aware=True)
+    versioned_file = versioned_bucket_name + '/versioned_file2'
+    for i in range(1200):
+        with s3.open(versioned_file, 'wb') as fo:
+            fo.write(b'1')
+    versions = s3.object_version_info(versioned_file)
+    assert len(versions) == 1200
+
+
+def test_versions_unaware(s3):
+    versioned_file = versioned_bucket_name + '/versioned_file3'
+    s3 = S3FileSystem(anon=False, version_aware=False)
+    with s3.open(versioned_file, 'wb') as fo:
+        fo.write(b'1')
+    with s3.open(versioned_file, 'wb') as fo:
+        fo.write(b'2')
+
+    with s3.open(versioned_file) as fo:
+        assert fo.version_id is None
+        assert fo.read() == b'2'
+
+    with pytest.raises(ValueError):
+        with s3.open(versioned_file, version_id='0'):
+            fo.read()
