@@ -5,6 +5,7 @@ import os
 import tempfile
 import shutil
 import sys
+import re
 
 PY2 = sys.version_info[0] == 2
 PY3 = sys.version_info[0] == 3
@@ -69,14 +70,12 @@ def seek_delimiter(file, delimiter, blocksize):
         last = full[-len(delimiter):]
 
 
-
 def read_block(f, offset, length, delimiter=None):
     """ Read a block of bytes from a file
 
     Parameters
     ----------
-    fn: string
-        Path to filename on S3
+    f: file
     offset: int
         Byte offset to start read
     length: int
@@ -119,8 +118,8 @@ def read_block(f, offset, length, delimiter=None):
         length = end - start
 
     f.seek(offset)
-    bytes = f.read(length)
-    return bytes
+    b = f.read(length)
+    return b
 
 
 def raises(exc, lamda):
@@ -187,4 +186,4 @@ class SSEParams(object):
         self.SSEKMSKeyId = sse_kms_key_id
 
     def to_kwargs(self):
-        return {k: v for k, v in self.__dict__.items() if v is not None}
\ No newline at end of file
+        return {k: v for k, v in self.__dict__.items() if v is not None}
